.286
;MIDIEMU.ASM (C)1993 Johannes Bjerregaard. All rights reserved.
;--------------------------------------------------------------
;This module contains hardware-independent MIDI emulation functions which
;work in conjunction with a hardware-specific driver.
;
;This version of the MIDIEMU driver is meant to be linked with the other
;single segment model module "ADLIB.ASM" to provide the direct interface
;between the two driver parts. In a later version, these will be separated.
;
;It is important to link the MIDIEMU module BEFORE the ADLIB module. The
;MIDIEMU module provides the interface functions to the user.
;The functions of the ADLIB module will be called automatically from MIDIEMU
;at the appropriate times.
;
;Note that no variables should be public. Those that are are only public for
;debugging purposes. Public procedures may be removed in later versions
;where the modules are separatable.
;
;Note: The function processallsoundsoffmsg is not implemented in the assembler
;      version.
;

include version.inc
include adlib.inc

META_TRACKEND      equ 2fh                     ;Meta-event: end of track

include smfpcmds.inc
include emucmds.inc
include midimsg.inc
include smfptune.inc

IFDEF STANDALONE
.model tiny
ELSE
.model small,c
ENDIF

global midiemu_entry_point :far                ;global for debugging purposes
global hw_entry_point      :far
global hw_emuinit          :proc
global hw_loadtune         :proc
global hw_noteon           :proc
global hw_drumon           :proc
global hw_noteoff          :proc
global hw_pbchange         :proc
global hw_setport          :proc
global hw_setpbsens        :proc
global hw_setmodvalue      :proc
global hw_setpbvalue       :proc
global hw_isr              :proc
global fsmf_isr            :proc
global fsmf_loadtune       :proc
global hwvins              :byte
global vcop                :byte

.code

;----------------------------------------------------------------------------
;Procedure  :  midiemu_entry_point
;Purpose    :  Call a function numbered by the value in register BX
;Parameters :  BX = function number (MIDIEMU_xxxxx). Other registers vary
;              depending on function.
;
;Returns    :  -
;Remarks    :  This is the only address which is guaranteed not to change.
;              It must be located at CS:0.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
midiemu_entry_point proc  far
            pushf
            pusha
            push  ds
            push  es

IFNDEF STANDALONE
            push  ax
            mov   ax,@data
            mov   ds,ax
            pop   ax
ELSE
            push  cs
            pop   ds
ENDIF
            mov   es,tunedataseg
            shl   bx,1                  ;BX to word index
            mov   bx,[emu_func_table+bx];get address word of desired function
            call  bx

IFNDEF STANDALONE
;            call  testwrite
ENDIF

            pop   es
            pop   ds
            popa
            popf
            ret
midiemu_entry_point endp

.data
dataseg_id  label byte
            db    'SMFP MIDI Emulation Module V1.0'
            db    '(C)1993 Johannes Bjerregaard.'

emu_func_table  label word
            dw    offset cs:emu_emuinit
            dw    offset cs:emu_loadtune
            dw    offset cs:emu_putdat
            dw    offset cs:emu_isr
            dw    offset cs:fsmf_loadtune
            dw    offset cs:fsmf_isr
            dw    offset cs:emu_stoptune
IFDEF ENABLESFX
            dw    offset cs:sfx_setsegment
            dw    offset cs:sfx_playsound
            dw    offset cs:sfx_clearallsounds
IFDEF REMAP_SFX
            dw    offset cs:sfx_playsfx
ELSE
            dw    ?
ENDIF
ELSE
            dw    ?
            dw    ?
            dw    ?
            dw    ?
ENDIF

tunedataseg label word  ;segment of mfp
            dw    ?
hwvtime     label word   ;There's a copy of this variable in both of MIDIEMU
                         ;and ADLIB, to save time while keeping the modules
                         ;separatable. All other shared variables are made
                         ;accessible between the modules by providing
                         ;special functions to change each variable.

            dw    NUMHWVOICES dup (0)
chnlvolumevalue label byte
            db    16 dup (100)
chnlexpressionvalue label byte
            db    16 dup (100)
loopflag    label byte  ;set by emu_loadtune
            db    ?

;*******start of data initialized to 0
zeroinitstart label unknown

program     label byte  ;current active program for channels
            db    16 dup (?)
playing     label byte  ;set while tune is playing
            db    ?
midicmd     label byte  ;static local of emuputdat ...
            db    ?
cmd         label byte
            db    ?
cc          label byte
            db    ?
expected    label byte
            db    ?
databyte    label byte
            db    ?
chnl        label byte  ;... until here
            db    ?
modvalue    label byte
            db    16 dup (?)
chnlpbvalue label word
            dw    16 dup (?)
hold1       label byte
            db    16 dup (?)
chnlref     label byte
            db    NUMHWVOICES dup (?)
voicepriority label byte
            db    NUMHWVOICES dup (?)
channelnumvoice label byte
            db    16 dup (0)

zeroinitend label unknown

zeroinitlength equ zeroinitend-zeroinitstart



;*******start of data initialized to 0ffh

ffinitstart label unknown

voiceref    label byte
            db    1000h dup (?) ;16*128*2
freevoice   label byte
            db    NUMHWVOICES dup (0ffh)

ffinitend   label unknown
ffinitlength equ ffinitend-ffinitstart

;
;These are local temporary variables used in EMUPUTDAT
;

valueready  label word
            dw    ?
value       label word
            dw    ?

;
;These are procedure-local temporary variables
;(should have been located on stack)
;

prog        label word
            dw    ?
insnum      label byte
            db    ?
vel         label word
            dw    ?
hwv         label word
            dw    ?
vref        label word
            dw    ?
fndinsnum   label byte
            db    ?


;
;channelpriority and channelmaxvoice will be taken from
;the active tune, so they need not be defined in the data segment.
;

rpnlsb      label byte   ;LSB of registered parameter number (MSB ignored)
            db    ?

note        label byte
            db    ?
velocity    label byte
            db    ?
.code

;----------------------------------------------------------------------------
;Procedure  :  findfreevoice
;Purpose    :  Search for a voice in the destination hardware which is not
;              currently in use.
;Parameters :  BL = preferred instrument number of free voice.
;Returns    :  On success, AL = voice number (0 to NUMHWVOICES-1)
;              On failure, AL = NUMHWVOICES
;Remarks    :  Condition 1.
;              For time optimization reasons, the routine does not anymore
;              always return the voice which has been unused for the longest
;              period of time. A higher prioritized parameter is now the
;              instrument which was last used to play the free voice. This
;              option can be disabled by not defining the constant
;              INSTRUMENT_SEARCH in the header file VERSION.INC.
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
findfreevoice proc
            mov   al,NUMHWVOICES
            xor   ah,ah
            mov   dh,bl                ;DH = preferred instrument
            mov   dl,0ffh              ;and mask
            mov   cx,ax                ;count
            xor   bx,bx                ;voice word index
            xor   si,si                ;voice byte index
            xor   di,di                ;hitime
            mov   fndinsnum,bl         ;clear found instrument number
@nextvce:
            cmp   [freevoice+si],dl    ;DL = 0ffh
            jnz   @notfree

IFDEF INSTRUMENT_SEARCH
            cmp   [hwvins+si],dh       ;is this the preferred instrument
            jz    @insok

            ;if not preferred instrument, check if another voice with
            ;the preferred instrument has already been found.

            test  fndinsnum,dl
            jnz   @notfree             ;if so, don't use this instrument
ENDIF

@insok:

            cmp   [hwvtime+bx],di      ;longer than hitime?
            jb    @notfree             ;no
IFDEF INSTRUMENT_SEARCH
            inc   fndinsnum
ENDIF
            mov   ax,si
            mov   di,[hwvtime+bx]
@notfree:
            inc   bx
            inc   bx
            inc   si
            loop  @nextvce
@retvalue:
            ret
findfreevoice endp



;----------------------------------------------------------------------------
;Procedure  :  noteoffpatch
;Purpose    :  Note off the specified voice. If the voice is part of a patch
;              which consists of two instruments, the second instrument will
;              be noted off too.
;Parameters :  AH = MIDI channel (0-9)     AL = Voice number
;Returns    :  -
;Remarks    :  Preserves register AX (MIDI channel and voice number)
;Registers
; destroyed :  BX,CX,DX,ES,SI,DI (ES loads SMFP info block)
;                                (CX and DI destroyed by hw_noteoff)
;----------------------------------------------------------------------------
noteoffpatch proc
            push  ax
            call  hw_noteoff
            pop   ax
            mov   bl,al
            xor   bh,bh
            mov   dl,[freevoice+bx]
            xor   dh,dh
            shl   dl,1                 ;key * 2
            mov   dh,ah                ;channel * 256
            mov   si,dx
            mov   bl,[voiceref+si]     ;BL = possible second voice
            cmp   bl,al                ;is this the voice parameter?
            jnz   @vcnotparam

            ;
            ;The voice found in the voiceref was the same as the one being
            ;passed as parameter. Try the other one.
            ;

            mov   bl,[voiceref+si+1]   ;possible second voice

@vcnotparam:
            cmp   bl,0ffh              ;unused?
            jz    @vcnosecond

            push  bx
            push  ax
            push  si
            mov   al,bl
            call  hw_noteoff           ;note off second voice
            pop   si
            pop   ax
            pop   bx

            mov   [freevoice+bx],0ffh  ;mark 2nd voice as free
            mov   [chnlref+bx],10h     ;remove channel reference
            mov   bl,ah
            dec   [channelnumvoice+bx]

@vcnosecond:
            mov   word ptr [voiceref+si],0ffffh; voice refrence, both voices
            mov   bl,al                ;fix note on info for 1st voice
            mov   [freevoice+bx],0ffh
            mov   [chnlref+bx],10h
            mov   bl,ah
            dec   [channelnumvoice+bx]
            ret
noteoffpatch endp







;----------------------------------------------------------------------------
;Procedure  :  replacevoice
;Purpose    :  Search for a voice in the destination hardware using the
;              specified channel which is currently playing a key that is
;              less important than the specified key.
;Parameters :  BL = channel, AH = key
;Returns    :  On success, AL = voice number (0 to NUMHWVOICES-1)
;              On failure, AL = NUMHWVOICES
;Remarks    :  Condition 2.
;Registers
; destroyed :  AH,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
replacevoice proc
            xor   bh,bh

            ;If the maxvoice parameter of the specified channel is higher
            ;than the current number of voices on this channel, return
            ;immediately without taking a voice; to steal a voice from an-
            ;other channel.

            mov   al,NUMHWVOICES
            mov   dl,[channelnumvoice+bx]
            cmp   dl,es:[channelmaxvoice+bx]
            jb    dontreplace

            mov   cx,ax                ;count
            xor   ch,ch
            xor   si,si                ;voice byte index
            mov   dh,bl

            cmp   bl,9                 ;channel 9 (drums) ?

            jnz   notchnl9

            ;find among the voices on channel 9 the one which has the lowest
            ;priority.

            mov   bl,ah                ;key to BL
            xor   bh,bh
c9nextvce:
            mov   dl,es:[drumpriority+bx] ;DL = priority of new key

            cmp   [voicepriority+si],dl;current lower than AH?
            jae   c9notlower           ;no, keep voice on
            mov   ax,si                ;remember voice
            mov   dl,[voicepriority+si];and lowest priority so far
c9notlower:
            inc   si                   ;inc voice byte index
            loop  c9nextvce            ;for all voice on channel 9
            jmp   c9done               ;channel 9 done

            ;
            ;This is how the best channel is found for replacement when
            ;the channel is not 9. The voice on the specified channel which
            ;were on the longest time is replaced.
            ;
            mov   dh,bl
notchnl9:
            mov   dl,0ffh              ;and mask
            xor   bx,bx                ;BX = voice word index
            xor   di,di                ;hitime
nc9nextvce:
            cmp   [chnlref+si],dh      ;is it the wanted channel?
            jnz   @notchnl

            cmp   [hwvtime+bx],di      ;longer than hitime?
            jb    @notchnl             ;no

            mov   ax,si                ;remember voice number
            mov   di,[hwvtime+bx]      ;and longest time so far
@notchnl:
            inc   bx
            inc   bx
            inc   si
            loop  nc9nextvce
dontreplace:
c9done:
            cmp   al,NUMHWVOICES       ;found one?
            jz    @replace_notfound    ;no replacement found

            mov   ah,dh                ;channel to ah
            call  noteoffpatch         ;causes no harm to AL
@replace_notfound:
            ret
replacevoice endp





;----------------------------------------------------------------------------
;Procedure  :  stealvoice
;Purpose    :  Take a voice from a different channel than the passed one.
;              The voice taken is the one which has the lowest priority of
;              its channel, or, if a drum channel, of its key priority.
;Parameters :  BL = channel
;Returns    :  On success, AL = voice number (0 to NUMHWVOICES-1)
;              On failure, AL = NUMHWVOICES
;Remarks    :  Condition 2.
;Registers
; destroyed :  AH,BX,CX,DX,SI,DI,ES
;----------------------------------------------------------------------------
stealvoice  proc
            mov   ax,NUMHWVOICES
            mov   cx,ax
            xor   si,si                   ;index
            mov   dh,DRUM_MIN_PRIORITY    ;minpri
            xor   bx,bx
@st_nv:
            mov   bl,[chnlref+si]
            mov   dl,es:[channelmaxvoice+bx]
            cmp   [channelnumvoice+bx],dl
            jbe   @st_hi

            cmp   [voicepriority+si],dh   ;voicepri<minpri?
            jae   @st_hi

            ;
            ;remember the voice with lowest priority
            ;

            mov   ax,si                   ;remember voice
            mov   dh,[voicepriority+si]   ;remember priority value
            mov   di,bx                   ;remember channel
@st_hi:
            inc   si
            loop  @st_nv

            cmp   al,NUMHWVOICES          ;found any voice?
            jz    @st_no                  ;no voice found

            mov   bl,al                   ;BL==voice
            mov   ax,di                   ;DI==AX==channel
            mov   ah,al                   ;retreive channel
            mov   al,bl                   ;retreive voice
            call  noteoffpatch            ;ah=channel,al=voice
@st_no:
            ret
stealvoice  endp


;----------------------------------------------------------------------------
;Procedure  :  getvoice
;Purpose    :  Find a voice suitable for playing the specified key and
;              channel. Depending on different states, one of the following
;              three conditions may be successful:
;
;              1) If an unused voice exists, take it.
;
;              2) If the channel's voice limit has been exceeded, replace
;                 the less important of the voices already on on that channel.
;
;              3) If any other channel is playing a key which has a lower
;                 priority than the specified channel and key, replace it
;                 with this.
;
;Parameters :  AH = MIDI channel (0-9)     AL = Key number
;              BL = Instrument number
;Returns    :  AL = voice number or NUMHWVOICES on failure.
;Remarks    :  The functionality of this procedure is found in the procedures
;              findfreevoice (condition 1), replacevoice (condition 2) and
;              stealvoice (condition 3).
;Registers
; destroyed :  AX,BX,CX,DX,ES,SI,DI
;----------------------------------------------------------------------------
getvoice    proc
            push   ax
            call   findfreevoice             ;condition 1
            cmp    al,NUMHWVOICES
            jz     @gv_1
@gv_fnd:
            inc    sp                        ;adjust stack
            inc    sp
            ret
@gv_1:
            pop    ax
            push   ax
            mov    bl,ah                     ;BL = channel
            mov    ah,al                     ;AH = key
            call   replacevoice              ;condition 2
            cmp    al,NUMHWVOICES
            jnz    @gv_fnd

            pop    ax

            ;
            ;only if there is at least one voice in spare in the
            ;channelmaxvoice table, request the procedure stealvoice to
            ;brutally cut a note off another channel.
            ;

            mov   bl,ah
            xor   bh,bh
            mov   cl,es:[channelmaxvoice+bx]
            cmp   [channelnumvoice+bx],cl
            jae   @gv_norequest              ;enough already

            call  stealvoice                 ;expects BL = channel
            ret

@gv_norequest:
            mov   al,NUMHWVOICES
            ret
getvoice    endp




;----------------------------------------------------------------------------
;Procedure  :  processnotemsg
;Purpose    :  Process note on and note off messages.
;Parameters :  AL = MIDI channel (0-15)      AH = key
;              DL = velocity
;Returns    :  -
;Remarks    :  Differentation between note on and note off messages is
;              determined by a velocity of nonzero versus zero.
;Registers
; destroyed :  AH,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
processnotemsg proc
            and   dl,dl                ;velocity 0 (note off?)
            jnz   pn_noteon
            jmp   pn_noteoff

pn_noteon:
            ;
            ;Processing a note on event.
            ;

            ;get the program number to use. If channel == 9, ...

            cmp  al,9
            jnz  pn_not9

            ;... get the program through the drumins table with key as index

            mov  bl,ah
            xor  bh,bh
            mov  dh,es:[drumins+bx]
            jmp  pn_gotprogram

pn_not9:
            ;... otherwise, use default program for channel.

            mov  bl,al
            xor  bh,bh
            mov  dh,[program+bx]

pn_gotprogram:

            ;save the program number in local temporary
            ;the program number does not point to the instrument itself
            ;but is merely a first-order index into the two-dimentional table
            ;that contains the instrument numbers to play with the two
            ;"pseudo-operators".

            mov  byte ptr prog,dh

            ;
            ;Calculate a resulting volume of the key, taking into account
            ;the current values of velocity, channel volume and channel
            ;expression. A "true" MIDI emulator would not calculate one
            ;volume for each note on message but one for every volume
            ;MIDI message.
            ;

            mov   si,ax             ;save channel and key
            mov   bl,al             ;channel index (BH already 0)

            mov   al,[chnlvolumevalue+bx]
            mul   dl                ;channel volume * velocity

            mov   dl,[chnlexpressionvalue+bx]
            xor   dh,dh
            mul   dx

            ;shift DX:AX right 2*7 times.
            ;instead, shift DX:AX right two times and move DX to AX

            shl   ax,1
            rcl   dx,1
            shl   ax,1
            rcl   dx,1
            mov   ax,dx
            mov   byte ptr vel,al

            ;------------------------------------------
            ;Set up parameters for first voice of patch
            ;------------------------------------------

            mov   bl,byte ptr prog
            shl   bl,1
            mov   al,es:[patchtable+bx+0]       ;ins1

            cmp   al,0ffh                    ;undefined?
            jnz   t1_defined
            jmp   t1_undefined
t1_defined:
            mov   insnum,al
            ;
            ;if channel+note is already in use, noteoff the
            ;voice and resuse it ....
            ;

            ;first calculate index into voiceref.

            mov   bx,si                      ;restore channel+key to BX
            xchg  bh,bl                      ;channel mul 256
            shl   bl,1                       ;+ key mul 2
            mov   vref,bx                    ;save in local temp
            mov   al,[voiceref+bx]
            cmp   al,0ffh                    ;already in use?
            jz    t1_notinuse

            ;the note and channel was used, note off voice for reuse, and
            ;skip the initialization that is not nessecary.

            push  ax
            push  bx
            push  dx
            push  si
            call  hw_noteoff
            pop   si
            pop   dx
            pop   bx
            pop   ax

            mov   byte ptr hwv,al
            xor   ah,ah
            mov   di,ax                      ;DI = voice byte index

            jmp   t1_gotvoice
t1_notinuse:

            ;... otherwise, search for a voice that can be used.

            mov   ax,si                      ;restore key + channel
            xchg  al,ah                      ;prepare parameters
            push  si
            mov   bl,insnum                  ;BL = preferred instrument
            call  getvoice
            pop   si
            cmp   al,NUMHWVOICES             ;no voice found?
            jnz   pn_found1                  ;if not, exit
            jmp   pn_exit

pn_found1:
            mov   byte ptr hwv,al

            ;set up voice data initialization

            xor   ah,ah
            mov   di,ax                      ;DI = voice byte index
            mov   ax,si

            mov   [freevoice+di],ah          ;AH = key
            mov   bl,al
            xor   bh,bh
            inc   [channelnumvoice+bx]       ;BX = channel byte index

            ;following initialization is located differently from in the
            ;C implementation - take care.

            mov   [chnlref+di],al            ;set up channel reference

            mov   bx,vref
            mov   al,byte ptr hwv
            mov   [voiceref+bx],al

t1_gotvoice:
            ;
            ;Initialize settings for pitch bend and modulation whell with
            ;the current settings for the channel.
            ;

            mov   ax,si
            xor   ah,ah
            mov   bx,ax                      ;BX = channel byte index
            shl   bx,1                       ;word index
            mov   cx,[chnlpbvalue+bx]
            shr   bx,1                       ;back to byte index
            mov   ah,byte ptr hwv
            push  bx
            call  hw_setpbvalue
            pop   bx
;            mov   ah,byte ptr hwv
            mov   al,[modvalue+bx]
            push  bx
            call  hw_setmodvalue
            pop   bx

            ;
            ;set the priority of the new voice and note on.
            ;

            mov   ax,si
            cmp   al,9                       ;if channel is 9...
            jnz   t1_usechnlpri

            ;... use priority from drum key priority table ...

            mov  bl,ah
            xor  bh,bh
            mov  al,es:[drumpriority+bx]
            mov  [voicepriority+di],al

            ;note on

            mov  ch,es:[drumnote+bx]
            mov  al,insnum
            jmp  short t1_priset

t1_usechnlpri:
            ;... otherwise use default priority for channel.

            mov  al,es:[channelpriority+bx]
            mov  [voicepriority+di],al
            mov  ch,ah
            mov  al,insnum
t1_priset:
            mov  bx,di
            shl  bx,1
            mov  [hwvtime+bx],0
            mov  ah,byte ptr hwv
            mov  cl,byte ptr vel
            push si

            call hw_noteon

            pop  si
t1_undefined:

            ;-------------------------------------------
            ;Set up parameters for second voice of patch
            ;-------------------------------------------

            ;NOTE: this is an exact copy of the above lines of code except
            ;      for a "+1" a few places, so remember that any changes to
            ;      the code should be updated both places...

            mov   bl,byte ptr prog
            shl   bl,1
            mov   al,es:[patchtable+bx+1]       ;ins2

            cmp   al,0ffh                    ;undefined?
            jnz   t2_defined
            jmp   t2_undefined
t2_defined:
            mov   insnum,al
            ;
            ;if channel+note is already in use, noteoff the
            ;voice and resuse it ....
            ;

            ;first calculate index into voiceref.

            mov   bx,si                      ;restore channel+key to BX
            xchg  bh,bl                      ;channel mul 256
            shl   bl,1                       ;+ key mul 2
            mov   vref,bx                    ;save in local temp
            mov   al,[voiceref+bx+1]
            cmp   al,0ffh                    ;already in use?
            jz    t2_notinuse

            ;the note and channel was used, note off voice for reuse, and
            ;skip the initialization that is not nessecary.

            push  ax
            push  bx
            push  dx
            push  si
            call  hw_noteoff
            pop   si
            pop   dx
            pop   bx
            pop   ax

            mov   byte ptr hwv,al
            xor   ah,ah
            mov   di,ax                      ;DI = voice byte index

            jmp   t2_gotvoice
t2_notinuse:

            ;... otherwise, search for a voice that can be used.

            mov   ax,si                      ;restore key + channel
            xchg  al,ah                      ;prepare parameters
            push  si
            mov   bl,insnum                  ;BL = preferred instrument
            call  getvoice
            pop   si
            cmp   al,NUMHWVOICES             ;no voice found?
            jnz   pn_found2
            jmp   pn_exit                    ;if not, exit
pn_found2:
            mov   byte ptr hwv,al

            ;set up voice data initialization

            xor   ah,ah
            mov   di,ax                      ;DI = voice byte index
            mov   ax,si

            mov   [freevoice+di],ah          ;AH = key
            mov   bl,al
            xor   bh,bh
            inc   [channelnumvoice+bx]       ;BX = channel byte index

            ;following initialization is located differently from in the
            ;C implementation - take care.

            mov   [chnlref+di],al            ;set up channel reference

            mov   bx,vref
            mov   al,byte ptr hwv
            mov   [voiceref+bx+1],al

t2_gotvoice:
            ;
            ;Initialize settings for pitch bend and modulation whell with
            ;the current settings for the channel.
            ;

            mov   ax,si
            xor   ah,ah
            mov   bx,ax                      ;BX = channel byte index
            shl   bx,1                       ;word index
            mov   cx,[chnlpbvalue+bx]
            shr   bx,1                       ;back to byte index
            mov   ah,byte ptr hwv
            push  bx
            call  hw_setpbvalue
            pop   bx
;            mov   ah,byte ptr hwv
            mov   al,[modvalue+bx]
            push  bx
            call  hw_setmodvalue
            pop   bx

            ;
            ;set the priority of the new voice and note on.
            ;

            mov   ax,si
            cmp   al,9                       ;if channel is 9...
            jnz   t2_usechnlpri

            ;... use priority from drum key priority table ...

            mov  bl,ah
            xor  bh,bh
            mov  al,es:[drumpriority+bx]
            mov  [voicepriority+di],al

            ;note on

            mov  ch,es:[drumnote+bx]
            mov  al,insnum
            jmp  short t2_priset

t2_usechnlpri:
            ;... otherwise use default priority for channel.

            mov  al,es:[channelpriority+bx]
            mov  [voicepriority+di],al
            mov  ch,ah
            mov  al,insnum
t2_priset:
            mov  bx,di
            shl  bx,1
            mov  [hwvtime+bx],0
            mov  ah,byte ptr hwv
            mov  cl,byte ptr vel
            push si
            call hw_noteon
            pop  si
t2_undefined:
            jmp  pn_exit

pn_noteoff:
            ;a note off message has been received. If the note and channel
            ;is on, turn it off. It may not be on since the note on message
            ;may have been masked out.

            mov   si,ax                      ;save for later retrieval
            xor   ah,ah
            mov   di,ax                      ;DI = channel byte index
            mov   dh,0ffh                    ;0ffh is a convenient value
                                             ;to keep in a register at the
                                             ;moment

            ;check if the voice reference is valid

            mov   bx,si                      ;restore channel+key to BX
            xchg  bh,bl                      ;channel mul 256
            shl   bl,1                       ;+ key mul 2
            mov   dl,[voiceref+bx]
            cmp   dl,dh
            jz    no_noton1                  ;no, skip it

            mov   [voiceref+bx],dh

            dec   [channelnumvoice+di]

            mov   al,dl
@notchnl9:
            push  bx
            push  dx
            push  di
            call  hw_noteoff
            pop   di
            pop   dx

            mov   bl,dl
            xor   bh,bh
            mov   [freevoice+bx],dh

            pop   bx

no_noton1:
            mov   dl,[voiceref+bx+1]
            cmp   dl,dh
            jz    no_noton2                  ;not on

            mov   [voiceref+bx+1],dh
            mov   bl,dl
            xor   bh,bh
            mov   [freevoice+bx],dh
            mov   al,dl

            push  di
            call  hw_noteoff
            pop   di

            dec   [channelnumvoice+di]
no_noton2:
pn_exit:                                     ;exit from procedure
            ret
processnotemsg endp




;----------------------------------------------------------------------------
;Procedure  :  processpitchbendmsg
;Purpose    :  Process a pitch bend change MIDI message.
;Parameters :  BL = MIDI channel (0-15)
;              AL = pitch bend lsb         AH = pitch bend msb
;Returns    :  -
;Remarks    :  Each voice on the specified channel gets its pitch updated.
;Registers
; destroyed :  AH,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
processpitchbendmsg proc

            ;first, convert the lsb/msb value into an 14 bit value in AX

            shr   ah,1
            jnc   pb_n1
            or    al,80h
pb_n1:
            sub   ax,2000h               ;2000h is zero level

            ;find the voices which need to be updated

            xor   bh,bh
            shl   bl,1
            mov   [chnlpbvalue+bx],ax
            shr   bl,1

            mov   cx,NUMHWVOICES
            xor   si,si
pb_nv:
            cmp   [chnlref+si],bl        ;channel equal?
            jnz   pb_n2

            push  ax
            push  cx
            push  bx
            push  si

            mov   cx,ax
            mov   ax,si
            mov   ah,al
            call  hw_setpbvalue
            mov   al,ah
            call  hw_pbchange

            pop   si
            pop   bx
            pop   cx
            pop   ax
pb_n2:
            inc   si
            loop  pb_nv

            ret
processpitchbendmsg endp





;----------------------------------------------------------------------------
;Procedure  :  processmodulationmsg
;Purpose    :  Process a modulation control MIDI message
;Parameters :  BL = MIDI channel (0-15)
;              AL = new value of modulation control
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  BX,CX,DL,SI,DI
;----------------------------------------------------------------------------
processmodulationmsg proc

            ;find the voices which need to be updated

            mov   cx,NUMHWVOICES

            mov   dl,bl
            xor   bh,bh
            mov   [modvalue+bx],al
            xor   bl,bl
mm_nv:
            cmp   [chnlref+bx],dl        ;channel equal?
            jnz   mm_n2

            push  bx
            mov   ah,bl
            call  hw_setmodvalue
            pop   bx
mm_n2:
            inc   bx
            loop  mm_nv
            ret
processmodulationmsg endp



;----------------------------------------------------------------------------
;Procedure  :  processvolumemsg
;Purpose    :  Process a channel volume  MIDI message
;Parameters :  BL = MIDI channel (0-15)     AL = value
;Returns    :  -
;Remarks    :  Only stores the volume value in a module local variable
;Registers
; destroyed :  BH
;----------------------------------------------------------------------------
processvolumemsg proc
            xor   bh,bh
            mov   [chnlvolumevalue+bx],al
            ret
processvolumemsg endp




;----------------------------------------------------------------------------
;Procedure  :  processexpressionmsg
;Purpose    :  Process a channel expression  MIDI message
;Parameters :  BL = MIDI channel (0-15)     AL = value
;Returns    :  -
;Remarks    :  Only stores the expression value in a module local variable
;Registers
; destroyed :  BH
;----------------------------------------------------------------------------
processexpressionmsg proc
            xor   bh,bh
            mov   [chnlexpressionvalue+bx],al
            ret
processexpressionmsg endp


;----------------------------------------------------------------------------
;Procedure  :  processallnotesoffmsg
;Purpose    :  Process an all notes off MIDI message but turning all notes
;              off that are on on a specified MIDI channel.
;Parameters :  AL = MIDI channel (0-15)
;Returns    :  -
;Remarks    :
;Registers
; destroyed :  AH,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
processallnotesoffmsg proc
            mov   cx,NUMHWVOICES
            xor   bx,bx       ;byte index
@nextvce1:
            cmp   [chnlref+bx],al   ;using this channel?
            jnz   notchnl

            cmp   [freevoice+bx],0ffh;voice currently on?
            jnz   notchnl

            push  ax
            push  bx
            push  cx

            mov   al,bl             ;voice number to note off
            call  hw_noteoff

            pop   cx
            pop   bx
            pop   ax
notchnl:
            inc   bx
            loop  @nextvce1
            ret
processallnotesoffmsg endp



;----------------------------------------------------------------------------
;Procedure  :  processresetctrlsmsg
;Purpose    :  Process a reset all controllers MIDI message of a specied
;              channel:
;Parameters :  AL = MIDI channel (0-15)
;Returns    :  -
;Remarks    :
;Registers
; destroyed :  AX,BX
;----------------------------------------------------------------------------
processresetctrlsmsg proc
            mov   bl,al
            xor   bh,bh
            mov   [modvalue+bx],bh
            shl   bx,1
            mov   [chnlpbvalue+bx],0h
            ret
processresetctrlsmsg endp



;----------------------------------------------------------------------------
;Procedure  :  inittables
;Purpose    :  Initialize module local tables with default values.
;Parameters :  -
;Returns    :  -
;Remarks    :  This procedure should be called when a tune is initialized.
;Registers
; destroyed :  BH
;----------------------------------------------------------------------------
inittables  proc

            ;initialize channel default values

            mov   cx,16                     ;for each MIDI channel
            mov   al,100                    ;default value to store in
                                            ;volume and expression
            xor   bx,bx                     ;byte index
            xor   si,si                     ;word index
            xor   dx,dx                     ;pitch bend zero level
@it_nch:
            mov   [chnlvolumevalue+bx],al
            mov   [chnlexpressionvalue+bx],al
            mov   [chnlpbvalue+si],dx
            mov   [channelnumvoice+bx],dl

            inc   bx
            inc   si
            inc   si
            loop  @it_nch

            ;
            ;Fill the voiceref table with 0ffh's.
            ;

            cld
            mov   cx,(16*128*2)/2
            lea   di,voiceref
            push  es
            push  ds
            pop   es
            xor   ax,ax
            dec   ax
            rep   stosw
            pop   es

            ;initialize voice to key and voice to channel reference

            mov   cx,NUMHWVOICES
            xor   bx,bx
            xor   dl,dl
            mov   ah,16                      ;max. MIDI channel number +1
@it_nv:
            mov   [freevoice+bx],al          ;set to 0ffh
            mov   [chnlref+bx],ah
            mov   [voicepriority+bx],dl
            inc   bx
            loop  @it_nv

            ret
inittables  endp



;----------------------------------------------------------------------------
;Procedure  :  emu_resync
;Purpose    :  Resynchronize input stream with MIDI emulation module.
;Parameters :  -
;Returns    :  -
;Remarks    :  No special value is expected after calling this command.
;              It should only be called when errors occur in the module
;              that transmits MIDI bytes so it gets out of sync with this.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
emu_resync  proc
            mov   expected,EXP_ANYTHING
            ret
emu_resync  endp

;----------------------------------------------------------------------------
;Procedure  :  emu_putdat
;Purpose    :  Receive and process MIDI bytes from SMF reader.
;Parameters :  AL = MIDI data byte
;Returns    :  -
;Remarks    :  emu_resync only needs to be called in case of difficulties
;              with the SMF reader or the MIDI data themselves. This
;              function never gets out of sync.
;Registers
; destroyed :  all
;----------------------------------------------------------------------------
emu_putdat  proc
            mov   byte ptr value,al
            mov   byte ptr valueready,1

            cmp   expected,EXP_ANYTHING
            jz    expanything
            jmp   notanything
expanything:
            ;
            ;possible message types are:
            ;any message header (80h-0ffh)
            ;any message using running status (0-0ffh)
            ;

            ;check for 8 bit commands

            ;active sensing

            mov   valueready,0
            cmp   al,CMD_ACTIVE_SENSING
            jnz   notactivesensing
            jmp   exit_putdat

notactivesensing:

            ;exclusive begin

            cmp   al,CMD_EXCLUSIVE
            jnz   notexclusivebegin
            mov   expected,EXP_EXCLUSIVE_END
            jmp   exit_putdat

notexclusivebegin:

            ;events utilizing running status

            cmp   al,80h
            jae   notrunningstatus

            mov   valueready,1
            mov   ah,midicmd
            mov   cmd,ah
            call  cmdset
            jmp   @valrdy

notrunningstatus:
            mov   midicmd,al
            mov   cmd,al

            ;retrieve channel from active command

            mov   ah,al
            and   al,0fh
            mov   chnl,al

            ;mask channel off command

            and   ah,0f0h
            mov   midicmd,ah

cmdset:
            ;switch midicmd

            cmp   ah,CMD_NOTE
            jz    ep_cmdnote
            cmp   ah,CMD_NOTEOFF
            jz    ep_cmdnoteoff
            cmp   ah,CMD_CC
            jz    ep_cmdcc
            cmp   ah,CMD_POLY_KEY_PRES
            jz    ep_cmdpoly
            cmp   ah,CMD_PROGRAM_CHANGE
            jz    ep_cmdpc
            cmp   ah,CMD_CHANNEL_PRES
            jz    ep_cmdcp
            cmp   ah,CMD_PITCH_BEND
            jz    ep_cmdpb
            jmp   ep_done

ep_cmdnote:
            mov   expected,EXP_NOTE
            jmp   ep_done
ep_cmdnoteoff:
            mov   expected,EXP_NOTEOFF_NOTE
            jmp   ep_done
ep_cmdcc:
            mov   expected,EXP_CC_MSG
            jmp   ep_done
ep_cmdpoly:
            mov   expected,EXP_NOTE
            jmp   ep_done
ep_cmdpc:
            mov   expected,EXP_PROGRAM
            jmp   ep_done
ep_cmdcp:
            mov   expected,EXP_CHANNEL_PRES
            jmp   ep_done
ep_cmdpb:
            mov   expected,EXP_PITCH_BEND_LSB

ep_done:
            ret
notanything:

            cmp   valueready,0
            jnz   @valrdy
            jmp   @valnotrdy
@valrdy:
            ;switch (expected)

            mov   al,expected
            cmp   al,EXP_NOTE
            jz    prc_note
            cmp   al,EXP_VELOCITY
            jz    prc_velocity
            cmp   al,EXP_NOTEOFF_NOTE
            jz    prc_noteoff_note
            cmp   al,EXP_NOTEOFF_VELOCITY
            jz    prc_noteoff_velocity
            cmp   al,EXP_EXCLUSIVE_END
            jz    prc_exclusive_end
            cmp   al,EXP_CC_MSG
            jnz   not_prc_cc_msg
            jmp   prc_cc_msg
not_prc_cc_msg:
            cmp   al,EXP_CC_VALUE
            jnz   not_prc_cc_value
            jmp   prc_cc_value
not_prc_cc_value:
            cmp   al,EXP_PROGRAM
            jnz   not_prc_program
            jmp   prc_program
not_prc_program:
            cmp   al,EXP_PITCH_BEND_LSB
            jnz   not_prc_pitch_bend_lsb
            jmp   prc_pitch_bend_lsb
not_prc_pitch_bend_lsb:
            cmp   al,EXP_PITCH_BEND_MSB
            jnz   not_prc_pitch_bend_msb
            jmp   prc_pitch_bend_msb
not_prc_pitch_bend_msb:
            ;unknown value of expected
            mov   expected,EXP_ANYTHING
            jmp   prc_done

prc_note:
            mov   al,byte ptr value
            mov   note,al
            mov   expected,EXP_VELOCITY
            jmp   prc_done

prc_velocity:
            mov   al,byte ptr value
            mov   velocity,al
            mov   expected,EXP_ANYTHING

            mov   dl,al
            mov   ah,note
            mov   al,chnl
            call  processnotemsg
            jmp   prc_done

prc_noteoff_note:
            mov   al,byte ptr value
            mov   note,al
            mov   expected,EXP_NOTEOFF_VELOCITY
            jmp   prc_done

prc_noteoff_velocity:
            mov   al,byte ptr value
            mov   velocity,al
            mov   expected,EXP_ANYTHING

            mov   al,chnl
            mov   ah,note
            xor   dl,dl    ;note off : velocity eqs 0
            call  processnotemsg
            jmp   prc_done

prc_exclusive_end:
            cmp   value,CMD_EXCLUSIVE_END
            jz    @exc_end
            mov   expected,EXP_ANYTHING
            jmp   prc_done
@exc_end:
            mov   expected,EXP_ANYTHING
            jmp   prc_done

prc_cc_msg:
            mov   al,byte ptr value
            mov   cc,al
            mov   expected,EXP_CC_VALUE
            jmp   prc_done

prc_program:
            mov   bl,chnl
            xor   bh,bh
            mov   al,byte ptr value
            mov   [program+bx],al
            mov   expected,EXP_ANYTHING
            jmp   prc_done

prc_pitch_bend_lsb:
            mov   al,byte ptr value
            mov   databyte,al
            mov   expected,EXP_PITCH_BEND_MSB
            jmp   prc_done

prc_pitch_bend_msb:
            mov   expected,EXP_ANYTHING
            mov   al,databyte
            mov   ah,byte ptr value
            mov   bl,chnl
            call  processpitchbendmsg
            jmp   prc_done

            ;processing CC (control change) and CM (channel mode) values
            ;requires yet another switch statement...

prc_cc_value:
            ;switch(cc)

            mov   al,cc
            cmp   al,CC_MODULATION
            jz    pcc_mod
            cmp   al,CC_VOLUME
            jz    pcc_volume
            cmp   al,CC_EXPRESSION
            jz    pcc_expression
;            cmp   al,CC_RPN_LSB
;            jz    pcc_rpnlsb
;            cmp   al,CC_DATA_ENTRY_MSB
;            jz    pcc_dem
            cmp   al,CM_ALL_NOTES_OFF
            jz    pcc_ano
            cmp   al,CM_ALL_SOUNDS_OFF
            jz    pcc_aso
            cmp   al,CM_RESET_ALL_CTRLS
            jz    pcc_resetctrls
;            cmp   al,CC_BANK_SELECT
;            jz    pcc_bank

            ;default
            mov   expected,EXP_ANYTHING
            jmp   prc_done

pcc_ano:
            mov   al,chnl
            call  processallnotesoffmsg
            mov   expected,EXP_ANYTHING
            jmp   prc_done
pcc_aso:
            mov   al,chnl
            call  processallnotesoffmsg   ;      __NOTE__ this is not correct
            mov   expected,EXP_ANYTHING
            jmp   prc_done

pcc_resetctrls:
            mov   al,chnl
            call  processresetctrlsmsg
            mov   expected,EXP_ANYTHING
            jmp   prc_done
pcc_mod:
            mov   al,byte ptr value
            mov   bl,chnl
            call  processmodulationmsg
            mov   expected,EXP_ANYTHING
            jmp   prc_done
pcc_volume:
            mov   al,byte ptr value
            mov   bl,chnl
            call  processvolumemsg
            mov   expected,EXP_ANYTHING
            jmp   prc_done
pcc_expression:
            mov   al,byte ptr value
            mov   bl,chnl
            call  processexpressionmsg
            mov   expected,EXP_ANYTHING
            jmp   prc_done
pcc_rpnlsb:
            mov   al,byte ptr value
            mov   rpnlsb,al
            mov   expected,EXP_ANYTHING
            jmp   prc_done
;pcc_dem:                      ;data entry MSB
;            cmp   rpnlsb,0    ;only pitch bend sens RPN supported
;            jz    rpn_pbsens
;            mov   expected,EXP_ANYTHING
;            jmp   prc_done
;rpn_pbsens:
;            mov   al,byte ptr value
;            call  hw_setpbsens
;            mov   expected,EXP_ANYTHING
;            jmp   prc_done
pcc_bank:
            ;UNMASK the following four lines only if GS support is needed and
            ;supported by the hardware driver.

;            mov   al,value
;            mov   bl,chnl
;            xor   bh,bh
;            mov   [gsbank+bx],al
            mov   expected,EXP_ANYTHING
prc_done:
@valnotrdy:

exit_putdat:
            ret
emu_putdat  endp






;----------------------------------------------------------------------------
;Procedure  :  emu_loadtune
;Purpose    :  Set up patch data for a tune to play.
;Parameters :  AX = segment of SMFP data.
;Returns    :  CF set = error, clear = success. Returned from hw_loadtune.
;Remarks    :
;Registers
; destroyed :  BH
;----------------------------------------------------------------------------
emu_loadtune proc
            push  ax
            call  inittables
            call  emu_resync
            pop   ax
            push  es
            mov   es,ax
            mov   tunedataseg,ax
            mov   bh,es:[tuneloop]
            mov   loopflag,bh
            call  hw_loadtune           ;returns error flag in CF
            pop   es
            ret
emu_loadtune endp



;
emu_emuinit proc
        mov    cx,zeroinitlength
        xor    bx,bx
        lea    di,zeroinitstart
        push   es
        push   ds
        pop    es
        xor    ax,ax
        shr    cx,1
        jnc    cxin
        stosb
cxin:
        rep    stosw

        dec    ax
        lea    di,ffinitstart
        mov    cx,ffinitlength
        shr    cx,1
        jnc    cxin2
        stosb
cxin2:
        rep    stosw
        pop    es
        call   hw_emuinit
        ret
emu_emuinit endp



emu_isr proc
        mov   cx,9
        mov   bx,16
bxl1:
        inc   [hwvtime+bx]
        dec   bx
        dec   bx
        loop  bxl1
        mov   bx,11
        call  hw_entry_point

IFDEF ENABLESFX
        call  sfx_isr
ENDIF
        ret
emu_isr endp




.data

midiTimer label dword            ;incremented at 70Hz, HI dword specifies
                                 ;the time in SMF time units since tune start.
        dw    ?                  ;Low order (fractional) word of MIDI clock
        dd    ?                  ;High order dword of MIDI clock

midiFile label word              ;segment address of MIDI file
        dw    ?
midiPos  label dword             ;current segment/offset pointer of MIDI file
        dd    ?
midiTime label dword
        dd    ?                  ;time of next MIDI event, relative to
                                 ;previous MIDI event.

midiNext label dword             ;Absolute time of next MIDI event.
        dd    ?

midiTimerAdd label dword         ;value to add to midiTimer every frame to
                                 ;keep midiTimer by value as close to the
                                 ;time resolution specified by the SMF.
        dd    359512             ;This value is calculated using the
                                 ;following formula:
                                 ;
                                 ;          (bpm SHL 16)*tpq
                                 ;    m = --------------------
                                 ;               fps*60
                                 ;
                                 ;bpm is number of beats per minute.
                                 ;60 indicates number of seconds per minute.
                                 ;fps is the number of frames per second.
                                 ;tpq is the number of ticks per quarter.
                                 ;
                                 ;The bpm is shifted left 16 times so that
                                 ;the high order part of the timer value can
                                 ;serve as an approximate of the time
                                 ;resolution specified by the SMF.

                                 ;This formula could be calculated run-time
                                 ;if the routine is expected to be run at
                                 ;other frequencies later on or support SMFs
                                 ;running at other speeds. The value of
                                 ;359512 is calculated from a frame rate
                                 ;(fps) of 70, a bpm of 120 and a tpq of 192.

runStat  label byte              ;Current running status.
       dw ?
frameEventCount label byte       ;Number of events processed in current frame
       db ?
.code


;----------------------------------------------------------------------------
;Procedure  :  smf_readtime
;Purpose    :  Reads the time attribute of the next MIDI event
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :
;----------------------------------------------------------------------------
smf_readtime proc
        push   ds
        pop    es
        lds    si,midiPos
        xor    bx,bx     ;lo
        xor    dx,dx     ;hi
        xor    ah,ah

@rt_nxbyt:
        lodsb
        add    bx,ax
        adc    dx,0
        test   al,80h
        jz     @rt_endtime

        mov    dh,dl
        mov    dl,bh
        mov    bh,bl
        shr    dx,1
        rcr    bx,1
        jmp    short @rt_nxbyt

@rt_endtime:
        mov    word ptr es:midiPos+2,ds
        push   es
        pop    ds
        mov    word ptr midiPos,si

        mov    word ptr midiTime,bx
        mov    word ptr midiTime+2,dx

        add    word ptr midiNext,bx
        adc    word ptr midiNext+2,dx

        ret
smf_readtime endp



putMIDIByte macro    ;saves all registers, calls emu_putdata, restores
                     ;all registers
        pusha
        push  ds
        push  es
        pop   ds
        mov   es,tunedataseg
        call  emu_putdat
        push  ds
        pop   es
        pop   ds
        popa
endm




IFNDEF STANDALONE
testwrite   proc
            push  es
            push  cx
            push  di
            push  si
            push  ax
            cld

            mov   cx,80*5
            mov   ax,0b800h
            mov   es,ax
            lea   si,dataseg_id
            mov   di,160*20
            mov   ah,7fh
vbix:
            lodsb
            stosw
            loop  vbix

            pop   ax
            pop   si
            pop   di
            pop   cx
            pop   es
            ret
testwrite   endp
ENDIF




;----------------------------------------------------------------------------
;Procedure  :  fsmf_isr
;Purpose    :  Called at frequency of 70Hz, dispatches MIDI events to the
;              emu_putdat procedure at the proper times.
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :
;----------------------------------------------------------------------------
fsmf_isr proc
        mov    frameEventCount,0

        ;increment MIDI clock

        cmp    playing,0
        jnz    @isplaying

        jmp    smf_isr_noread
@isplaying:

        mov    ax,word ptr midiTimerAdd
        add    word ptr midiTimer,ax
        mov    ax,word ptr midiTimerAdd+2
        adc    word ptr midiTimer+2,ax
        adc    word ptr midiTimer+4,0

        ;if MIDI clock exceeded time of next MIDI event...

        mov    ax,word ptr midiNext
        sub    ax,word ptr midiTimer+2
        mov    ax,word ptr midiNext+2
        sbb    ax,word ptr midiTimer+4
        jc     @clk_exceed
        jmp    smf_isr_noread

@clk_exceed:

        ;... read the next event from the SMF stream

        push   ds                ;Save DS...
        pop    es                ;...In ES

        lds    si,midiPos

@clk_zero:
        lodsb
        cmp    al,0ffh           ;a meta-event?
        jnz    @noMeta

        ;Meta events are ignored. The EOT should cause the SMF to restart if
        ;specified by the MFP.

        lodsw                    ;meta-event type in AL, length in AH
        mov    bl,al
        mov    al,ah
        xor    ah,ah
        add    si,ax             ;add length of meta-event to midiPos
        cmp    bl,META_TRACKEND  ;end of track?
        jnz    @not_trackend

        ;restart smf?

        xor    ax,ax
        cmp    es:loopflag,al
        jnz    @looptune         ;yes

        ;stop playing tune

        push   es
        pop    ds
        mov    playing,al        ;stop playing tune
        jmp    smf_isr_noread

@looptune:
        mov    word ptr es:midiPos,16h
        mov    si,16h

        mov    word ptr es:midiTime,ax
        mov    word ptr es:midiTime+2,ax
        mov    word ptr es:midiNext,ax
        mov    word ptr es:midiNext+2,ax

        mov    word ptr es:midiTimer,ax
        mov    word ptr es:midiTimer+2,ax
        mov    word ptr es:midiTimer+4,ax

@not_trackend:
        jmp    @done_event

@noMeta:
        cmp    al,80h            ;is the event using running status?
        jb     @runStat          ;yes, skip

        mov    es:runStat,al     ;save running status

        putMIDIByte

        and    al,0f0h
        cmp    al,0c0h           ;C0 or D0 are only 2 byte events
        jz     @by2evt
        cmp    al,0d0h           ;C0 or D0 are only 2 byte events
        jz     @by2evt

        lodsb                    ;get next byte

        putMIDIByte
@by2evt:
        lodsb

        putMIDIByte

        jmp    short @done_event

@runStat:
        putMIDIByte

        mov    al,es:runStat
        and    al,0f0h
        cmp    al,0c0h
        jz     @done_event
        cmp    al,0d0h
        jz     @done_event

        lodsb
        putMIDIByte

@done_event:
@skip_event:
        ;read the time of the next event and, if the time is 0, go back
        ;and process it. Otherwise, exit.

        xor    bx,bx     ;lo
        xor    dx,dx     ;hi

@si_nxbyt:
        xor    ah,ah
        lodsb
        add    bx,ax
        adc    dx,0
        test   al,80h
        jz     @si_endtime

        and    bl,7fh
        mov    dh,dl
        mov    dl,bh
        mov    bh,bl
        xor    bl,bl
        shr    dx,1
        rcr    bx,1
        jmp    short @si_nxbyt

@si_endtime:
        and    ax,ax
        jnz    @nt_set
        and    dx,dx
        jnz    @nt_set

        ;The time is 0, go back and process the next event

        inc    es:frameEventCount
        cmp    es:frameEventCount,MAXFRAMEEVENTCOUNT
        jae    @nt_set

        jmp    @clk_zero

@nt_set:
        ;save the new position of the midi file pointer and the event time.

        mov    word ptr es:midiPos+2,ds
        push   es
        pop    ds
        mov    word ptr midiPos,si

        mov    word ptr midiTime,bx
        mov    word ptr midiTime+2,dx

        add    word ptr midiNext,bx
        adc    word ptr midiNext+2,dx

smf_isr_noread:
        call   emu_isr

;        call   testwrite
        ret
fsmf_isr endp

;----------------------------------------------------------------------------
;Procedure  :  fsmf_loadtune
;Purpose    :  Set up data for a tune to play.
;Parameters :  AX = segment of SMF
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :
;----------------------------------------------------------------------------
fsmf_loadtune proc

        push   ax
        call   emu_emuinit
        pop    ax
        mov    midiFile,ax
        mov    word ptr midiPos+2,ax
        mov    playing,1

        xor    ax,ax
        mov    word ptr midiPos,16h ;wind SMF to beginning of event data

        mov    word ptr midiNext,ax
        mov    word ptr midiNext+2,ax

        mov    word ptr midiTimer,ax
        mov    word ptr midiTimer+2,ax
        mov    word ptr midiTimer+4,ax

        ;Read time attribute of first event of SMF.

        call   smf_readtime
        ret

fsmf_loadtune endp


;----------------------------------------------------------------------------
;Procedure  :  emu_stoptune
;Purpose    :  Abort playing the active tune.
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  CX,AL,BX,DX
;----------------------------------------------------------------------------
emu_stoptune proc
;        pushf
;        cli
;        mov    playing,0
;        call   emu_resync

;        mov    cx,16
;@@es_nch:

;        mov    al,cl
;        add    al,0afh ; = 0b0h+channel
;        putMIDIbyte
;        mov    al,07bh
;        putMIDIbyte
;        xor    al,al
;        putMIDIbyte

;        loop   @@es_nch

;        popf
        call   emu_emuinit
        ret
emu_stoptune endp









IFDEF ENABLESFX

;****************************************************************************
;****************************************************************************
;**  .    ...   .    .  .  .   .   .      .   .  .               .        .**
;** .  .   .*****  **** .**  **.**. **.***** .    . ***** **  **.  .       **
;** .    . |**|**.** |**|** |**|** |**|** |**  - -- **-- -**- **           **
;**    .   |**| .|** |**|**.|**|***|**|**.|** . .- -**  - -****  . .       **
;**. .  . .|*****|** |**|**.|**|******|** |**  .  --****  --**   .  .      **
;**      .  . |**|**.|**|** |**|** ***|** |**  .   -**- -- ****    .      .**
;**  .   .. **|**|** |**|** |**|** |**|** |**.   - -**    **--**  . .  .   **
;**.    .  |*****.|**** .|**** |**.|**|*****   -----**   -** -**.    .  .  **
;**  . .   .   .      . .   .         .  .    .  ----  -- - ---    .       **
;****************************************************************************
;****************************************************************************

global   sfxnum           :byte

.data

sfxseg   label word       ;Segment of SFX data
         dw    ?

sfxnum   label byte                   ;Number of sound effect playing in each
                                      ;hardware voice.
         db    NUMHWVOICES dup (0ffh) ;Voices not in use for sound fx are
                                      ;reset to 0ffh.
sfxtime  label word                   ;Elapsed time of currently playing fx
         dw    NUMHWVOICES dup (0)    ;used to time keyoff and sfx shutdown
fnumtime label word                   ;Time left until next f-num packet
         dw    NUMHWVOICES dup (0)    ;0 = no time left
fnumptr  label word                   ;offset of current fnum packet
         dw    NUMHWVOICES dup (0)    ;
lvtime   label byte                   ;Time left until next level packet
         db    NUMHWVOICES dup (0)    ;0 = no time left
lvptr    label word                   ;offset of current lv packet
         dw    NUMHWVOICES dup (0)    ;
fnum     label word                   ;f-num of voices
         dw    NUMHWVOICES dup (0)    ;lower 16 bits are always <400h
sfxlv    label byte                   ;level of voices
         db    NUMHWVOICES dup (0)
sfxkon   label byte                   ;keyon bit of sfx used by kon/off
         db    NUMHWVOICES dup (0)    ;timing
.code

;----------------------------------------------------------------------------
;AdlOutProc: subroutine version of adl_out macro - purpose is to save space at
;locations where speed is no problem.
;----------------------------------------------------------------------------
adloutproc   proc
             adl_out
             ret
adloutproc   endp


;----------------------------------------------------------------------------
;Procedure  :  sfx_clearallsounds
;Purpose    :  Clears all voices used by sound effects
;Parameters :  AX = segment of sound effects data.
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  AX,CX,SI
;----------------------------------------------------------------------------
sfx_clearallsounds proc
            ;all voices which sfxnum number is not 0ffh are cleared in
            ;hardware and reset to 0ffh.

            mov    cx,NUMHWVOICES
            xor    si,si
@@ntv:
            cmp    [sfxnum+si],0ffh                  ;voice in use for fx?
            jz     @@ntff

            mov    [sfxnum+si],0ffh                  ;mark as unused
            mov    [freevoice+si],0ffh

            mov   ax,si                              ;one reg per voice
            add   al,0b0h                            ;reset KON
            call  adloutproc

            mov   ax,0ff80h                          ;reset sustain + release
            add   al,[vcop+si]
            call  adloutproc                         ;operator 2
            add   al,3                               ;
            call  adloutproc                         ;operator 1
            sub   al,20h                             ;reset attack + decay
            call  adloutproc                         ;operator 1
            sub   al,3
            call  adloutproc                         ;operator 2
@@ntff:
            inc    si
            loop   @@ntv
            ret
sfx_clearallsounds endp




;----------------------------------------------------------------------------
;Procedure  :  sfx_requestvoice
;Purpose    :  Returns a hardware voice number to use for playing a sound
;              effect. If no voices are currently available, the one with
;              the lowest priority is taken. Because the functions getvoice,
;              replacevoice and stealvoice require the new voice to be
;              using a specific channel and instrument, they cannot be used.
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
sfx_requestvoice proc
            call   findfreevoice

            ret
sfx_requestvoice endp

;----------------------------------------------------------------------------
;Procedure  :  sfx_setsegment
;Purpose    :  Initialize sfx player by setting the segment of the sfx data.
;              At the same time, all playing sounds are shut up, since
;              otherwise unpredictable sounds might occur.
;Parameters :  AX = segment of sound effects data.
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
sfx_setsegment proc
            mov    sfxseg,ax
            call   sfx_clearallsounds
            ret
sfx_setsegment endp

IFDEF REMAP_SFX

.data
sfx_map     label byte
            db    0,-1
            db    1,-1
            db    2,-1
            db    3,-1
            db    4,-1
            db    5,6
            db    7,8
            db    9,-1
            db    10,-1
            db    11,-1
            db    12,-1
            db    13,-1
            db    14,-1
            db    15,-1
            db    16,-1
            db    17,-1
            db    18,-1
            db    19,-1
            db    20,-1
            db    21,-1
            db    22,-1
            db    23,-1
            db    24,-1
            db    25,-1
            db    26,-1
            db    27,-1
            db    28,-1
            db    29,-1
            db    30,-1
.code
;----------------------------------------------------------------------------
;Procedure  :  sfx_playsfx
;Purpose    :  Plays a sound effect by remapping the parameter value into
;              two sounds and dispatching them to sfx_playsound.
;Parameters :  AX = sound effect number, by table above
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
sfx_playsfx proc
            mov   bx,ax
            shl   bx,1
            mov   ax,word ptr [bx+sfx_map]
            cmp   al,-1
            jz    @@nolo
            push  ax
            xor   ah,ah
            call  sfx_playsound
            pop   ax
@@nolo:
            mov   al,ah
            xor   ah,ah
            cmp   al,-1
            jz    @@nohi
            call  sfx_playsound
@@nohi:
            ret
sfx_playsfx endp
ENDIF

;----------------------------------------------------------------------------
;Procedure  :  sfx_playsound
;Purpose    :  start playing a sound effect
;Parameters :  AX = sound effect number
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  CX,AL,BX,DX
;----------------------------------------------------------------------------
sfx_playsound proc
            push   ax
IFDEF INSTRUMENT_SEARCH
            mov    bl,-1            ;dummy preferred instrument
ENDIF
            call   getvoice
            mov    bl,al
            cmp    bl,NUMHWVOICES
            jnz    @@sps_vf

            ;no free voice exists, search for one

            mov    bl,-1            ;dummy channel for new voice
            call   stealvoice
            mov    bl,al
@@sps_vf:
            pop    ax

            ;a voice number is now available in BL.

            cmp    bl,NUMHWVOICES   ;this can happen if all voices are active
            jnz    @@isavail        ;and have a priority >= DRUM_MIN_PRIORITY

            jmp    @@noavail
@@isavail:
            xor    bh,bh
            mov    [bx+sfxnum],al
            mov    [bx+freevoice],128;invalid note
            mov    [bx+voicepriority],080h
            mov    [bx+hwvins],0ffh;invalidate instrument for hw_isr

            ;
            ;clear voice
            ;

            mov   ax,0ff80h                          ;reset sustain + release
            add   al,[vcop+bx]
            adl_out                                  ;operator 2
            add   al,3                               ;
            adl_out                                  ;operator 1
            sub   al,20h                             ;reset attack + decay
            adl_out                                  ;operator 1
            sub   al,3
            adl_out                                  ;operator 2
            mov   ax,bx                              ;one reg per voice
            add   al,0b0h                            ;reset KON
            adl_out

            ;
            ;set up initial parameters for sound
            ;

            mov   ax,sfxseg
            mov   es,ax
            mov   al,[bx+sfxnum]
            xor   ah,ah
            mov   si,ax                              ;SI = sfxnum
            shl   si,1
            mov   si,es:[si]                         ;SI = offset of
                                                     ;     sfx structure
            ;BX = hw voice byte index
            ;SI = sfx structure offset into sfx segment

            ;operator 1

            mov   al,[bx+vcop]
            add   al,20h
            mov   ah,es:[si+sfxh_multi+0] ;multi etc.
            adl_out

            add   al,20h
            mov   ah,es:[si+sfxh_level+0] ;tl etc.
            adl_out

            add   al,20h
            mov   ah,es:[si+sfxh_ad+0]    ;a/d
            adl_out

            add   al,20h
            mov   ah,es:[si+sfxh_sr+0]    ;s/r
            adl_out

            add   al,60h
            mov   ah,es:[si+sfxh_wfm+0]   ;waveform
            adl_out

            ;operator 2

            sub   al,0c0h-3               ;back to register 20h+operator inc
            mov   ah,es:[si+sfxh_multi+1] ;multi etc.
            adl_out

            add   al,40h
            mov   ah,es:[si+sfxh_ad+1]    ;a/d
            adl_out

            add   al,20h
            mov   ah,es:[si+sfxh_sr+1]    ;s/r
            adl_out

            add   al,60h
            mov   ah,es:[si+sfxh_wfm+1]   ;waveform
            adl_out

            ;operator common

            mov   ax,bx
            add   al,0c0h
            mov   ah,es:[si+sfxh_fbc]     ;feedback+connection
            adl_out

            mov   di,bx
            shl   di,1

            ;DI = hw voice word index

            xor   ax,ax
            mov   [sfxtime+di],ax
            mov   [fnumtime+di],ax
            mov   [lvtime+bx],al
            mov   [sfxkon+bx],20h
            mov   ax,es:[si+sfxh_lvdata]
            sub   ax,2
            mov   [lvptr+di],ax
            mov   ax,es:[si+sfxh_fqdata]
            sub   ax,4
            mov   [fnumptr+di],ax
@@noavail:
            ret
sfx_playsound endp







;----------------------------------------------------------------------------
;Procedure  :  sfx_isr
;Purpose    :  real time update of sound effects
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :
;----------------------------------------------------------------------------
sfx_isr     proc
            push  bp
            push  es
            mov   ax,sfxseg
            mov   es,ax
            mov   si,NUMHWVOICES-1              ;SI = voice byte index
@@si_nvce:
            cmp   [sfxnum+si],0ffh              ;sfx enabled?
            jnz   @@si_ison
            jmp   @@si_noton
@@si_ison:

            mov   bl,[sfxnum+si]
            xor   bh,bh
            shl   bx,1
            mov   di,es:[bx]                    ;DI = sound offset in sfxseg

            mov   bx,si
            shl   bx,1                          ;BX = voice word index

            inc   [sfxtime+bx]
            mov   ax,[sfxtime+bx]
            cmp   ax,es:[di+sfxh_len2]          ;end of sound?
            jb    @@si_ny1

            mov   [sfxnum+si],0ffh              ;reset sound to unused
            mov   [freevoice+si],0ffh

            ;reset sound registers to stop playing sound effect

            mov   ax,0ff80h
            add   al,[vcop+si]
            adl_out

            add   al,3
            adl_out                             ;sustain/release, op2

            mov   ax,si
            xor   ah,ah
            add   al,0b0h
            adl_out                             ;reset keyon
            jmp   @@si_noton

@@si_ny1:
            cmp   ax,es:[di+sfxh_len1]          ;end of keyON?
            jb    @@si_ny2

            ;reset keyon bit to zero

            mov   [sfxkon+si],0

@@si_ny2:
            cmp   [lvtime+si],0                 ;rdy 4 new lv packet?
            jnz   @@si_slvp

            ;increment level data packet

            add   [lvptr+bx],2                  ;increment packet
@@sl_rld:
            mov   bp,[lvptr+bx]                 ;BP = lv packet base index
            mov   ax,es:[bp]                    ;AL = level
            cmp   ax,0ffffh                     ;loop?
            jnz   @@si_lnl

            mov   ax,es:[bp+2]                  ;load new start point
            mov   [lvptr+bx],ax                 ;update
            jmp   @@sl_rld                      ;reload
@@si_lnl:
            ;set up time count of the new packet

            mov   ax,es:[bp]

            mov   [lvtime+si],ah
            and   [lvtime+si],07fh
            jmp   @@si_lvndc
@@si_slvp:
            dec   [lvtime+si]                   ;decrement level packet time

            mov   bp,[lvptr+bx]                 ;BP = lv packet base index
            mov   ax,es:[bp]                    ;AL = level
@@si_lvndc:
            test  ah,80h                        ;is value absolute?
            jnz   @@si_lvnr                     ;not relative

            ;her behandles relative levels

            mov   dx,ax
            xor   dh,dh

            mov   cl,[sfxlv+si]
            xor   ch,ch
            add   dx,cx
            jns   @@si_lvp                      ;positive
            xor   dx,dx                         ;reset to max. level
            jmp   @@si_lvok
@@si_lvp:
            cmp   dl,40h                        ;overflow?
            jb    @@si_lvok
            mov   dl,3fh
            jmp   @@si_lvok

@@si_lvnr:  ;her behandles ikke-relative levels

            mov   [sfxlv+si],al
            jmp   @@si_ssdl
@@si_lvok:
            mov   [sfxlv+si],dl
@@si_ssdl:
            ;time to take a look at f-num update

            cmp   [fnumtime+bx],0               ;end of f-num packet?
            jnz   @@fn_ne

            ;increment f-num packet

            add   [fnumptr+bx],4                ;2 words per packet
            mov   bp,[fnumptr+bx]
            mov   ax,es:[bp+2]
            and   ax,07fffh
            mov   [fnumtime+bx],ax

            mov   ax,es:[bp]                    ;data value
            cmp   ax,0ffffh                     ;loop?
            jnz   @@si_fnnl

            ;reset fnumptr

            mov   ax,es:[bp+2]
            mov   [fnumptr+bx],ax
            mov   bp,ax
            mov   ax,es:[bp+2]
            and   ax,07fffh
            mov   [fnumtime+bx],ax
            mov   ax,es:[bp]                    ;load with new data value
            jmp   @@si_fnnl

@@fn_ne:
            dec   [fnumtime+bx]
            mov   bp,[fnumptr+bx]
            mov   ax,es:[bp]

@@si_fnnl:
            test  word ptr es:[bp+2],8000h      ;absolute if set
            jnz   @@si_fnumrdy

            ;process relative fnum

            mov   cx,[fnum+bx]
            add   cx,ax
            mov   ax,cx
            and   cx,3ffh
            cmp   cx,688
            jb    @@si_fnl
            add   ax,680
@@si_fnl:
            cmp   cx,344
            ja    @@si_fng
            sub   ax,680
@@si_fng:
@@si_fnumrdy:
            mov   [fnum+bx],ax

            ;update hardware with the latest changes

            mov   al,[vcop+si]
            add   al,40h+3                     ;select the carrier operator
                                               ;and the KSL/TL register
            mov   ah,[sfxlv+si]
            adl_out

            mov   ax,si
            mov   ah,byte ptr [fnum+bx]
            add   al,0a0h
            adl_out

            add   al,10h
            mov   ah,byte ptr [fnum+bx+1]
            and   ah,1fh
            or    ah,[sfxkon+si]
            adl_out

@@si_noton:
            dec   si
            cmp   si,0ffffh
            jz    @@si_nd
            jmp   @@si_nvce
@@si_nd:
            pop   es
            pop   bp
            ret
sfx_isr     endp

ENDIF ;ENABLESFX

end midiemu_entry_point
